"""
    AbstractPerturbed{parallel} <: AbstractOptimizationLayer

Differentiable perturbation of a black box optimizer.

The parameter `parallel` is a boolean value, equal to true if the perturbations are run in parallel.

# Available implementations

- [`PerturbedAdditive`](@ref)
- [`PerturbedMultiplicative`](@ref)

These two subtypes share the following fields:

- `maximizer`: black box optimizer
- `ε`: magnitude of the perturbation
- `nb_samples::Int`: number of random samples for Monte-Carlo computations
- `rng::AbstractRNG`: random number generator
- `seed::Union{Nothing,Int}`: random seed
"""
abstract type AbstractPerturbed{parallel} <: AbstractOptimizationLayer end

"""
    sample_perturbations(perturbed::AbstractPerturbed, θ)

Draw random perturbations `Z` which will be applied to the objective direction `θ`.
"""
function sample_perturbations(perturbed::AbstractPerturbed, θ::AbstractArray)
    (; rng, seed, nb_samples) = perturbed
    seed!(rng, seed)
    Z_samples = [randn(rng, size(θ)) for _ in 1:nb_samples]
    return Z_samples
end

function compute_atoms(
    perturbed::AbstractPerturbed{false},
    θ::AbstractArray,
    Z_samples::Vector{<:AbstractArray};
    kwargs...,
)
    return [perturb_and_optimize(perturbed, θ, Z; kwargs...) for Z in Z_samples]
end

function compute_atoms(
    perturbed::AbstractPerturbed{true},
    θ::AbstractArray,
    Z_samples::Vector{<:AbstractArray};
    kwargs...,
)
    return ThreadsX.map(Z -> perturb_and_optimize(perturbed, θ, Z; kwargs...), Z_samples)
end

function compute_probability_distribution(
    perturbed::AbstractPerturbed,
    θ::AbstractArray,
    Z_samples::Vector{<:AbstractArray};
    kwargs...,
)
    atoms = compute_atoms(perturbed, θ, Z_samples; kwargs...)
    weights = ones(length(atoms)) ./ length(atoms)
    probadist = FixedAtomsProbabilityDistribution(atoms, weights)
    return probadist
end

"""
    compute_probability_distribution(perturbed::AbstractPerturbed, θ; kwargs...)

Turn random perturbations of `θ` into a distribution on polytope vertices.

Keyword arguments are passed to the underlying linear maximizer.
"""
function compute_probability_distribution(
    perturbed::AbstractPerturbed, θ::AbstractArray; kwargs...
)
    Z_samples = sample_perturbations(perturbed, θ)
    return compute_probability_distribution(perturbed, θ, Z_samples; kwargs...)
end

"""
    (perturbed::AbstractPerturbed)(θ; kwargs...)

Apply `compute_probability_distribution(perturbed, θ; kwargs...)` and return the expectation.
"""
function (perturbed::AbstractPerturbed)(θ::AbstractArray; kwargs...)
    probadist = compute_probability_distribution(perturbed, θ; kwargs...)
    return compute_expectation(probadist)
end
